#include "main.h"
#include "stm32f4xx_hal.h"
#include "stm32f4xx_nucleo.h"

#include "FreeRTOS.h"
#include "task.h"
#include "timers.h"
#include "queue.h"
#include "semphr.h"
#include "event_groups.h"
#include <string.h>

volatile uint8_t UART_RxBuffer[UART_BufferSize];
volatile uint8_t UART_TxBuffer[UART_BufferSize];
static char dataOut[MAX_BUF_SIZE];
const uint32_t Usart_BaudRate = 115200;
UART_HandleTypeDef UartHandle;

void SystemClock_Config(void);
void USARTConfig(void);

/* Tasks -----------------------------------------------------------*/
void vSenderTask(void const * argument);
void vReceiverTask(void const * argument);

/* Declare a variable of type QueueHandle_t.
 * This is used to store the handle to the queue that is accessed by all three tasks. */
QueueHandle_t xQueue;

int main( void )
{
	HAL_Init();
	SystemClock_Config();
	BSP_LED_Init(LED2);
	USARTConfig();

    /* The queue is created to hold a maximum of 5 values,
     * each of which is large enough to hold a variable of type int32_t. */
    xQueue = xQueueCreate(5, sizeof(uint32_t ));
    if( xQueue != NULL ) {
        /* Create two instances of the task that will send to the queue.
         * The task parameter is used to pass the value that the task will write to the queue, so one task will continuously write 100
         * to the queue while the other task will continuously write 200 to the queue.
         *  Both tasks are created at priority 1. */
        xTaskCreate((void *)vSenderTask, "Sender1", configMINIMAL_STACK_SIZE, (void*) 100, 1, NULL );
        xTaskCreate((void *)vSenderTask, "Sender2", configMINIMAL_STACK_SIZE, (void* ) 200, 1, NULL);

        /* Create the task that will read from the queue.
         * The task is created with priority 2, so above the priority of the sender tasks. */
        xTaskCreate((void *)vReceiverTask, "Receiver", configMINIMAL_STACK_SIZE, NULL, 2, NULL );

        /* Start the scheduler so the created tasks start executing. */
        vTaskStartScheduler();
    } else {
        /* The queue could not be created. */
    }

    while (1) {
  	  BSP_LED_Toggle(LED2);
  	  HAL_Delay(100);
    }
}

void vSenderTask( void const *pvParameters )
{
    uint32_t UlValueToSend;
    BaseType_t xStatus;
    /* Two instances of this task are created so the value that is sent to the queue is passed in through the task parameter.
     *  This way, each instance can use a different value.
     *  The queue was created to hold values of type int32_t, so cast the parameter to the required type. */
    UlValueToSend = (uint32_t) pvParameters;
    /* As per most tasks, this task is implemented within an infinite loop. */
    while(1) {
        /* Send the value to the queue. The first parameter is the queue to which data is being sent.
         *  The second parameter is the address of the data to be sent, in this case the address of lValueToSend.
         *  The third parameter is the Block time, the time the task should be kept in the Blocked state to wait for space to become available on the queue should the queue already be full.
         *  In this case a block time is not specified because the queue should never contain more than one item, and therefore never be full. */
        xStatus = xQueueSendToBack(xQueue, &UlValueToSend, 0);
        if(xStatus != pdPASS) {
            /* The send operation could not complete because the queue was full. This must be an error because the queue should never contain more than one item! */
        	snprintf( dataOut, MAX_BUF_SIZE, "Could not send to the queue.\r\n");
        	HAL_UART_Transmit( &UartHandle, ( uint8_t * )dataOut, strlen( dataOut ), 5000 );
        }
        vTaskDelay(500);
    }
}

void vReceiverTask( void const *pvParameters )
{
    /* Declare the variable that will hold the values received from the queue. */
    uint32_t UlReceivedValue;
    BaseType_t xStatus;
    const TickType_t xTicksToWait = pdMS_TO_TICKS( 1000 );
    /* This task is also defined within an infinite loop. */
    while(1) {
        /* This call should always find the queue empty because this task will immediately remove any data that is written to the queue. */
        if( uxQueueMessagesWaiting( xQueue ) != 0 ) {
        	snprintf( dataOut, MAX_BUF_SIZE, "Queue should have been empty!\r\n");
        	HAL_UART_Transmit( &UartHandle, ( uint8_t * )dataOut, strlen( dataOut ), 5000 );
        }
        /* Receive data from the queue.
         * The first parameter is the queue from which data is to be received.
         * The second parameter is the buffer into which the received data will be placed. In this case the buffer is simply the address of a variable that has the required size to hold the received data.
         * The last parameter is the block time, the maximum amount of time that the task will remain in the Blocked state to wait for data to be available should the queue already be empty.
         *  */
        xStatus = xQueueReceive(xQueue, &UlReceivedValue, xTicksToWait );
        if( xStatus == pdPASS ) {
            /* Data was successfully received from the queue, print out the received value. */
            snprintf( dataOut, MAX_BUF_SIZE, "Received:  %lu\r\n", UlReceivedValue );
            HAL_UART_Transmit( &UartHandle, ( uint8_t * )dataOut, strlen( dataOut ), 5000 );
        } else {
            /* Data was not received from the queue even after waiting for 100ms.
             * This must be an error because the sending tasks are free running and will be continuously writing to the queue. */
        	snprintf( dataOut, MAX_BUF_SIZE, "Could not receive from the queue.\r\n");
        	HAL_UART_Transmit( &UartHandle, ( uint8_t * )dataOut, strlen( dataOut ), 5000 );
        }
    }
}

void SystemClock_Config(void)
{

  RCC_OscInitTypeDef RCC_OscInitStruct;
  RCC_ClkInitTypeDef RCC_ClkInitStruct;

  __HAL_RCC_PWR_CLK_ENABLE();

  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE2);

  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = 16;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSI;
  RCC_OscInitStruct.PLL.PLLM = 16;
  RCC_OscInitStruct.PLL.PLLN = 336;
  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV4;
  RCC_OscInitStruct.PLL.PLLQ = 7;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    _Error_Handler(__FILE__, __LINE__);
  }

  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV2;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_2) != HAL_OK)
  {
    _Error_Handler(__FILE__, __LINE__);
  }

  HAL_SYSTICK_Config(HAL_RCC_GetHCLKFreq()/1000);
  HAL_SYSTICK_CLKSourceConfig(SYSTICK_CLKSOURCE_HCLK);
  HAL_NVIC_SetPriority(SysTick_IRQn, 15, 0);
}

void USARTConfig( void )
{

  GPIO_InitTypeDef GPIO_InitStruct;

  USARTx_TX_GPIO_CLK_ENABLE();
  USARTx_RX_GPIO_CLK_ENABLE();
  USARTx_CLK_ENABLE();

  GPIO_InitStruct.Pin       = USARTx_TX_PIN;
  GPIO_InitStruct.Mode      = GPIO_MODE_AF_PP;
  GPIO_InitStruct.Pull      = GPIO_NOPULL;
  GPIO_InitStruct.Speed     = GPIO_SPEED_FREQ_HIGH;
  GPIO_InitStruct.Alternate = USARTx_TX_AF;

  HAL_GPIO_Init(USARTx_TX_GPIO_PORT, &GPIO_InitStruct);

  GPIO_InitStruct.Pin       = USARTx_RX_PIN;
  GPIO_InitStruct.Alternate = USARTx_RX_AF;

  HAL_GPIO_Init( USARTx_RX_GPIO_PORT, &GPIO_InitStruct );

  UartHandle.Instance        = USARTx;
  UartHandle.Init.BaudRate   = Usart_BaudRate;
  UartHandle.Init.WordLength = UART_WORDLENGTH_8B;
  UartHandle.Init.StopBits   = UART_STOPBITS_1;
  UartHandle.Init.Parity     = UART_PARITY_NONE;
  UartHandle.Init.HwFlowCtl  = UART_HWCONTROL_NONE;
  UartHandle.Init.Mode       = UART_MODE_TX_RX;

  if ( HAL_UART_Init( &UartHandle ) != HAL_OK )
  {
    while( 1 );
  }

  UartHandle.pRxBuffPtr = ( uint8_t * )UART_RxBuffer;
  UartHandle.RxXferSize = UART_BufferSize;
  UartHandle.ErrorCode  = HAL_UART_ERROR_NONE;
}

void _Error_Handler(char *file, int line)
{
  while(1)
  {
  }
}

#ifdef  USE_FULL_ASSERT
void assert_failed(uint8_t* file, uint32_t line)
{ 
}
#endif
